const char rcsid[]=
"$Id: lav2axt.c,v 1.13 2002/08/30 02:33:50 schwartz Exp schwartz $";

/* lav2txt - produce alignments in the ".axt" format. */
//
// XXX - requires normalized lav
//

#define debug if(debug_flag)fprintf
#define USE_SEQCACHE

#define put(x) putchar_unlocked(x)

#include <stdio.h>
#include <signal.h>

#include "util.h"
#include "args.h"
#include "seq.h"
#include "mus_cdb.h"

static int suppress_autoalignment_flag = 0;
static int blockreverse_size = 0;
static int debug_flag = 0;
static int trim_contig_flag = 0;

extern char *revcomp(uchar *s, int len);

enum {
  BUF_SIZE=500
};

#define DEFAULT_REGION "genomic"
#define RC " (reverse complement)"

#define CDB_DIR "/genome/cache/mus/clipped"
char *cdb_dir = CDB_DIR;
char *fasta_dir = 0;

/* Gap-free segments of an alignment are stored in a linked list. */
typedef struct seg
{
	int beg1, beg2, end1, end2;
	struct seg *next;
} SEG;

/* Each alignment is stored as a header and a list of segments, and initially
* put into a linked list.  Later points to them are put in an array so they
* can be efficiently sorted */
typedef struct align {
	int beg1, beg2, end1, end2;
	int score;
	SEG *first;
	char *contig_name;
	int revcomp;
	struct align *next;
} ALIGN;

ALIGN *first_align = NULL;
enum { Hlen=1024, Blen = 4000000 };
unsigned char b1[Blen], b2[Blen];
unsigned char *s1;

// -- tiny cache
ALIGN *cachealign = 0; // for contig and strand
unsigned char *cacheseq = 0;
unsigned char *cacheseqr = 0;
int cachemisses = 0;

static char *blockreverse(char *s, int len);

unsigned char *lookaside(ALIGN *a)
{
	unsigned int len = 0;

	//fprintf(stderr, "contig_name: %s\n", a->contig_name);
	//fprintf(stderr, "revcomp: %d\n", a->revcomp);
	if ((cachealign == 0) ||
	    strcmp(a->contig_name, cachealign->contig_name)) {
		++cachemisses;
		//fprintf(stderr, "contig_name: !=\n");
		if (cacheseq) {
			free(cacheseq);
			free(cacheseqr);
		}
		cacheseq = mus_cdb_seq2(a->contig_name, &len);
		if (blockreverse_size) cacheseq = blockreverse(cacheseq,len);
		cacheseqr = revcomp(cacheseq, len);
	}
	cachealign = a;
	return a->revcomp ? cacheseqr : cacheseq;
}

// --- junk for reverse mouse 
#define SWAP(s,a,z) do{char t=s[a];s[a]=s[z];s[z]=t;}while(0)
static void reverse(char *s, int start, int end)
{
    while (start < end) {
        SWAP(s,start,end);
        start++, end--;
    }
}
static char *blockreverse(char *s, int len)
{
    int i;
    for (i=0; i<len; i+=blockreverse_size) {
        int j = MIN(len,i+blockreverse_size);
        reverse(s, i, j-1);
    }
    return s;
}
// --- 

void trim_contig_name(char *s)
{
    // map "/foo/bar/chr1.nib:99-999:xyz" to "chr1"
    char *p;
    if (trim_contig_flag == 0) return;

    p = strrchr(s, '/');
    if (p) { ++p; memmove(s,p,strlen(p)+1); }

    p = strchr(s, ':');
    if (p) *p = 0;

    p = strrchr(s, '.');
    if (p) *p = 0;
}

void gap_error(SEG *s)
{
    fatalf("no gaps between segments (%d->%d %d->%d) (%d %d)",
	    s->end1, s->next->beg1, s->end2, s->next->beg2,
	    s->next->beg1 - s->end1 - 1, s->next->beg2 - s->end2 - 1);
}

int get_lines(char *file)
{
	FILE *ap;
	char buf[BUF_SIZE], contig[BUF_SIZE], *p;
	ALIGN *newal = NULL, *prev = NULL;
	SEG *newseg, *lastseg = NULL;
	int nalign, revcomp;

	ap = ckopen(file, "r");
	if ((fgets(buf, BUF_SIZE, ap)) == NULL || strcmp(buf, "#:lav\n"))
		fatal("Not an alignment file.");
	revcomp = nalign = 0;
	contig[0] = '\0';
	while (fgets(buf, BUF_SIZE, ap)) {
		if (same_string(buf, "h {\n")) {
			fgets(buf, BUF_SIZE, ap);
			fgets(buf, BUF_SIZE, ap);
			if ((p = strchr(buf, '"')) == NULL)
				fatal("cannot find first paren");
			(void) strcpy(contig, p+1);
			if ((p = strchr(contig, '"')) == NULL)
				fatal("cannot find next paren");
			*p = '\0';
			if ((p = strstr(contig, RC)) != NULL) {
				*p = '\0';
				revcomp = 1;
			} else
				revcomp = 0;
		} else if (same_string(buf, "a {\n")) {
			++nalign;
			newal = ckalloc(sizeof(ALIGN));
			if (contig[0] == '\0')
				fatal("no contig name");
			newal->contig_name = copy_string(contig);
			trim_contig_name(newal->contig_name);
			newal->revcomp = revcomp;
			if (first_align == NULL)
				first_align = newal;
			else
				prev->next = newal;
			prev = newal;
			if  (fgets(buf, BUF_SIZE, ap) == NULL ||
			     sscanf(buf, "  s %d\n", &(newal->score)) != 1)
				fatalf("no score: %s", buf);
			if  (fgets(buf, BUF_SIZE, ap) == NULL ||
			     sscanf(buf, "  b %d %d", &(newal->beg1),
						      &(newal->beg2)) != 2) 
				fatalf("no beginning: %s", buf);
			if  (fgets(buf, BUF_SIZE, ap) == NULL ||
			     sscanf(buf, "  e %d %d", &(newal->end1),
						      &(newal->end2)) != 2) 
				fatalf("no ending: %s", buf);
			newal->first = NULL;
			while (fgets(buf, BUF_SIZE, ap) && buf[0] != '}') {
				newseg = ckalloc(sizeof(SEG));
				if (sscanf(buf, "  l %d %d %d %d",
				    &(newseg->beg1), &(newseg->beg2),
				    &(newseg->end1), &(newseg->end2)) != 4)
					fatalf("Cannot get segment: %s", buf);
				newseg->next = NULL;
				if (newal->first == NULL)
					newal->first = newseg;
				else
					lastseg->next = newseg;
				newseg->next = NULL;
				lastseg = newseg;
				
			}
		}
	}

	fclose(ap);
	return nalign;
}

int cmp_beg1(const void *a, const void *b)
{
	return (*((ALIGN **)a))->beg1 - (*((ALIGN **)b))->beg1;
}

int cmp_mus(const void *a, const void *b)
{
	ALIGN *aa = *((ALIGN **)a);
	ALIGN *bb = *((ALIGN **)b);
	int c;

	c = strcmp(aa->contig_name, bb->contig_name);
	return (c != 0) ? c : (aa->beg1 - bb->beg1);
}

// >/chromMixedNib.0729/chr1.nib:203000001-204010000:_

static char *ctok(char *buf, const char *p)
{
	// XXX - this is completely wrong
	char *q = buf;
	if (p == 0) { buf[0] = 0; return buf; }
	while (*p == '>') ++p;
	while (*p && isspace(*p)) ++p;
	while (*p && !isspace(*p) && *p != ':') *q++ = *p++;

	*q = 0;
	return buf;
}

//  XXX - Decide if this is what we mean.
int on_diagonal(ALIGN *a, const char *region)
{
	char buf[BUF_SIZE];
	if (a->revcomp == 0 &&
	    a->beg1 == a->beg2 &&
	    a->end1 == a->end2 &&
	    strcmp(region, ctok(buf,a->contig_name)) == 0)
		return 1;
	return 0;
}

void print_align(ALIGN *a, const char *region)
{
	static int nalign = 0;
	SEG *s;
	unsigned char *s2, *f1, *f2, *t1, *t2, *t;
	char buf[BUF_SIZE];

	if (suppress_autoalignment_flag)
	    if (on_diagonal(a,region))
		return;

	// num chr22 123 4567 contig9999 135 7890
	printf("%d %s %d %d %s %d %d %c %d\n",
		nalign,
		region, a->beg1, a->end1,
		ctok(buf,a->contig_name), a->beg2, a->end2,
		a->revcomp ? '-' : '+',
		a->score
	);

#ifdef USE_SEQCACHE
	s2 = lookaside(a);
#else
	s2 = mus_cdb_seq(a->contig_name);
	if (a->revcomp) do_revcomp(s2, strlen(s2));
#endif	

	t1 = b1;
	t2 = b2;
	f1 = s1 + a->beg1 - 1;
	f2 = s2 + a->beg2 - 1;

	if (a->first) debug(stderr, "\n");
	for (s = a->first; s; s = s->next) {
		int i, k = s->end1 - s->beg1 + 1;
		assert(s->beg1 <= s->end1+1);
		assert(s->beg2 <= s->end2+1);
		for (i = k; i > 0; --i) {
			// if (*f1 == *f2) ++nmatch;
			*t1++ = *f1++;
			*t2++ = *f2++;
		}
		debug(stderr, "= %d\n", k);
		if (s->next) {
			const int k1 = s->next->beg1 - s->end1 - 1;
			const int k2 = s->next->beg2 - s->end2 - 1;
			//assert((k1==0) ^ (k2==0));
			if (k1 > 0) { // XXX - SES - 1 or 0? was 1? causes gap_error.
				k = k1;
				debug(stderr, "+ %d\n", k);
				assert(t1-b1+k < Blen);
				assert(t2-b2+k < Blen);
				while (k-- > 0) {
					*t1++ = *f1++;
					*t2++ = '-';
				}
			} else {
				k = k2;
				debug(stderr, "- %d\n", k);
				assert(t1-b1+k < Blen);
				assert(t2-b2+k < Blen);
				if (k == 0) gap_error(s);
				while (k-- > 0) {
					*t1++ = '-';
					*t2++ = *f2++;
				}
			}
		}
	}

{
	int bp1, bp2;
	bp1 = bp2 = 0;
	for (t = b1; t < t1; ++t) {
		put(*t);
		if (*t != '-') ++bp1;
	}
	put('\n');
	for (t = b2; t < t2; ++t) {
		put(*t);
		if (*t != '-') ++bp2;
	}
	put('\n');

	// XXX -- we are missing trailing indels here

	assert(t1-b1 < Blen);
	assert(t1-b1 == t2-b2);
	if (a->end1-a->beg1+1 != bp1 || a->end2-a->beg2+1 != bp2)
	    fprintf(stderr, "%d ; %d ?= %d ; %d ?= %d\n",
		t2-b2,
		a->end1-a->beg1+1, bp1,
		a->end2-a->beg2+1, bp2);
	//assert(a->end1-a->beg1+1 == bp1);
	//assert(a->end2-a->beg2+1 == bp2);
}

	put('\n');
#ifndef USE_SEQCACHE
	free(s2);
#endif
	++nalign;
}

static void catcher(int sig)
{
	sig = 0;
	exit(1);
}

static char *basename(char *s)
{
    char *t = strrchr(s,'/');
    return t ? t+1 : s;
}

int main(int argc, char **argv)
{
	int i, n, n_align;
	ALIGN *a, **al;
	SEQ *sp;
	char *region, buf[200];

	signal(SIGINT, catcher);

	argv0 = basename(argv[0]);
	if (argc < 3)
	    fatal("usage: lav-file human-genomic-seq"
		  " [N=region-name] [D=cdb-directory] [F=fasta-directory]");

	ckargs("adrtDFN", argc, argv, 2);
        get_argval_pos('a', &suppress_autoalignment_flag, 0);
        get_argval_pos('t', &trim_contig_flag, 1);
        get_argval_pos('r', &blockreverse_size, 0);
        get_argval_pos('d', &debug_flag, 0);
	if (!get_cargval('N', &region)) {
		strcpy(buf, DEFAULT_REGION);
		region = buf;
	}
        if (!get_cargval('D', &cdb_dir))
                cdb_dir = CDB_DIR;
        if (!get_cargval('F', &fasta_dir))
                fasta_dir = 0;

	n_align = get_lines(argv[1]);

	if (n_align < 1)
		return 0;

	al = ckalloc(n_align*sizeof(ALIGN *));
	for (n = 0, a = first_align; a; a = a->next) {
		if (n >= n_align)
			fatal("Too many alignments.");
		al[n++] = a;
	}
	if (n != n_align)
		fatal("n != n_align");
	fprintf(stderr, "%d alignments\n", n_align);

	// Originally: sort alignments by beg1
	// qsort((void *)al, n_align, sizeof(ALIGN *), cmp_beg1);
	// No, sort by mouse, so the contig cache works 
	qsort((void *)al, n_align, sizeof(al[0]), cmp_mus);

	sp = seq_get(argv[2]);
	s1 = SEQ_CHARS(sp);
	mus_cdb_init2(cdb_dir,fasta_dir);
	for (i = 0; i < n_align; ++i) {
		print_align(al[i], region);
	}
	seq_close(sp);
	mus_cdb_terminate();

	return 0;
}
